/*
    player - WAV player
    Copyright (C) 2010 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#if defined(HAVE_SNDFILE)

int sw_player_kbd_func(struct subwin *sw, struct event *ev, int fw){
    if (!sdl) return 0;
    switch(kbd_action(KM_MAIN,ev)){
        case ACT_LEFT:
            if (!ssbd) break;
            MUTEX_LOCK(ssbd->seek);
            ssbd->seek -= 3000;
            MUTEX_UNLOCK(ssbd->seek);
        case ACT_RIGHT:
            if (!ssbd) break;
            MUTEX_LOCK(ssbd->seek);
            ssbd->seek += 1000;
            MUTEX_UNLOCK(ssbd->seek);
            return 1;
        case ACT_DOWN:
            if (!ssbd) break;
            MUTEX_LOCK(ssbd->seek);
            ssbd->seek -= 20000;
            MUTEX_UNLOCK(ssbd->seek);
        case ACT_UP:
            if (!ssbd) break;
            MUTEX_LOCK(ssbd->seek);
            ssbd->seek += 20000;
            MUTEX_UNLOCK(ssbd->seek);
            return 1;
    }

    return 0;
}

int sw_player_mouse_func(struct subwin *sw, struct event *ev, int fw){
//    if (!sdl) return 0;
    return 0;
}


void sw_player_redraw(struct subwin *sw){
	SDL_Rect area;
    int i, j, x, y0, y, c;

    dbg("sw_player_redraw(%d)\n", sw->gdirty);
#ifdef HAVE_SDL    
    if (!sdl) return;
    if (sw->h - PLAYER_H <= 0) return;

    fill_area(sw->x, sw->y, sw->w, sw->h - PLAYER_H , 0);
	area.x = 0;
	area.y = 0;
	area.w = sw->screen->w;
	area.h = sw->screen->h;
    SDL_FillRect(sw->screen, &area, makecol(40, 40, 40));

    //dbg("shapelen=%d channels=%d\n", sw->pl_shapelen, sw->pl_channels);
    if (sw->pl_pxlen == 0) return;
    c = sdl->green;
    sw_player_check_bounds(sw);
    int cursor = ssbd->played / sw->pl_pxlen;
    for (x = 0; x < sw->w * FONT_W; x++){
        i = x + sw->pl_ofs;
        if (i >= sw->pl_shapelen) break;
        if (i < 0) break;
        
        for (j = 0; j < sw->pl_channels; j++){
            y0 = sw->screen->h / (2 * sw->pl_channels);
            y = (sw->pl_shapebuf[i * sw->pl_channels + j] * sw->screen->h / 2) / ( sw->pl_channels * 256);
            if (cursor == i){
                y = sw->screen->h / (sw->pl_channels * 2);
                line(sw->screen, x, y0+y, x, y0-y, sdl->gr[15]);
            }else{
                line(sw->screen, x, y0+y, x, y0-y, c);
            }
        }
    }
    

#endif
}

void sw_player_check_bounds(struct subwin *sw){

    if (!sdl) return;
    if (sw->pl_pxlen == 0) return;

    int cursor = ssbd->played / sw->pl_pxlen;

    if (cursor - sw->pl_ofs < 0){
        sw->pl_ofs = cursor - sw->screen->w / 5;
    }

    if (cursor - sw->pl_ofs > (sw->screen->w * 4) / 5){
        sw->pl_ofs = cursor - (sw->screen->w * 4) / 5;
    }

}

void sw_player_raise(struct subwin *sw){
#ifdef HAVE_SDL
    sw->gdirty=1;      
#endif
}

void sw_player_free(struct subwin *sw){
    CONDGFREE(sw->pl_shapebuf);
}

#define MS2PX 20

void player_play(char *filename){
    struct subwin *sw;
    int i, j;
    SF_INFO sfinfo;
    SNDFILE *sf;
    short *pxbuf, *sh;
    double *sum;
    short *avg;
    short *max;
    int shapei = 0;

    for (i=0; i<gses->subwins->len; i++){
        sw = (struct subwin *)g_ptr_array_index(gses->subwins, i);
        if (sw->type != SWT_PLAYER) continue;
        sw_set_ontop(i);
        goto found;
    }
    sw = new_subwin(SWT_PLAYER, VTEXT(T_SW_PLAYER), NULL);
    sw_set_ontop(gses->subwins->len-1);

found:;    
    redraw_later();
	sw->gdirty=1;

    sw->pl_len = 0;
    sw->pl_channels = 1;
    memset (&sfinfo, 0, sizeof (sfinfo));
    sf = sf_open(filename, SFM_READ, &sfinfo);
    if (!sf) {
        log_addf("Can't play %s - %s", filename, sf_strerror (NULL));
        return;
    }

    sw->pl_len = sf_seek(sf, 0L, SEEK_END);
    if (sw->pl_len < 0) {
        log_addf("Can't seek to end %s - %s", filename, sf_strerror(sf));
        sf_close(sf);
        return;
    }

    if (sf_seek(sf, 0L, SEEK_SET) < 0) {
        log_addf("Can't seek to start %s - %s", filename, sf_strerror(sf));
        sf_close(sf);
        return;
    }

    sw->pl_channels = sfinfo.channels;
    sw->pl_rate = sfinfo.samplerate;
    sw->pl_pxlen = sw->pl_channels * sw->pl_rate * MS2PX / 1000;

    dbg("pl_len=%d pxlen=%d\n", sw->pl_len, sw->pl_pxlen);
    CONDGFREE(sw->pl_shapebuf);
    sw->pl_shapebuf = g_new0(char, sw->pl_channels * (sw->pl_len / sw->pl_pxlen)+1);
    sw->pl_shapelen = 0;

    pxbuf = g_new0(short, sw->pl_pxlen);
    sum = g_new0(double, sw->pl_channels);
    avg = g_new0(short, sw->pl_channels);
    max = g_new0(short, sw->pl_channels);
    shapei = 0;
    while(1){
        int readlen = sf_read_short(sf, pxbuf, sw->pl_pxlen);

        if (readlen <= 0) break;
        for (j = 0; j < sw->pl_channels; j++) {
            sum[j] = 0.0;
            max[j] = 0;
        }

        sh = pxbuf;
        for (i = 0; i < readlen; i++){
            for (j = 0; j < sw->pl_channels; j++){
                sum[j] += *sh;
                sh++;
            }
        }

        for (j = 0; j < sw->pl_channels; j++) avg[j] = (short)(sum[j] / readlen);

        sh = pxbuf;
        for (i = 0; i < readlen; i++){
            for (j = 0; j < sw->pl_channels; j++){
                int d = *sh - avg[j];
                if (d < 0) d = -d;
                if (d > max[j]) max[j] = d;
                sh++;
            }
        }

        for (j = 0; j < sw->pl_channels; j++) {
            int shape = 0;
          //  dbg("max[%d]=%5d ", j, max[j]);
            if (max[j] == 0) {
                shape = 0;
            }else{
                //shape = log(max[j]) * 256.0 / log(32768);
                shape = max[j] * 256.0 / 32768;
                if (shape > 255) shape = 255;
            }
           // dbg("shape=%3d ", shape);
            sw->pl_shapebuf[shapei] = shape;
            shapei++;
        }
        //dbg("\n");
        sw->pl_shapelen++;
    }
    sf_close(sf);
    g_free(pxbuf);
    g_free(sum);
    g_free(avg);
    g_free(max);
    
    ssbd_play_file(ssbd, filename);
    gses->icon=ssbd->playicon;
}

#else
int sw_player_kbd_func(struct subwin *sw, struct event *ev, int fw){
    return 0;
}

int sw_player_mouse_func(struct subwin *sw, struct event *ev, int fw){
    return 0;
}

void sw_player_redraw(struct subwin *sw){
}

void sw_player_check_bounds(struct subwin *sw){
}

void sw_player_raise(struct subwin *sw){
}

#endif
