
/*
    Tucnak - VHF contest log
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct qrvdb *qrvdb;


struct qrvdb *init_qrv(void){
    struct qrvdb *qrvdb;
    int i;

    qrvdb = g_new0(struct qrvdb, 1);
    init_list(qrvdb->list);
    qrvdb->cur=g_new0(struct qrv_item *, 'Z'-'A'+1);
    qrvdb->def_bands = 0;
    qrvdb->hash = g_hash_table_new(g_str_hash, g_str_equal);

    for (i=0; i<cfg->bands->len; i++){
        struct config_band *confb = g_ptr_array_index(cfg->bands, i);
        if (confb->qrv || confb->qrvnow) qrvdb->def_bands |= 1 << (toupper(confb->bandchar)-'A');
    }    

    return qrvdb;
}

void free_qrv_item(struct qrv_item *qi){
    CONDGFREE(qi->call);
    CONDGFREE(qi->wwl);
    CONDGFREE(qi->text);
    g_free(qi);
}

void clear_qrv(struct qrvdb *qrvdb){
    int i;
    struct subwin *sw;
    
    // minqrv is cleared also in maps_reload but we must be sure it will not point to freeed memory
    if (gses && gses->subwins){  // gses==NULL when terminating
        for (i=0;i<gses->subwins->len;i++){
            sw=(struct subwin *)g_ptr_array_index(gses->subwins, i);
            if (sw->type!=SWT_MAP) continue;
#ifdef HAVE_SDL            
            sw->minqrv=NULL;
#endif
        }
    }

    for (i=0; i<'Z'-'A'+1; i++) qrvdb->cur[i]=NULL;

    while(qrvdb->list.next != &qrvdb->list){
        struct qrv_item *qi;
        qi = qrvdb->list.next;
        del_from_list(qi);
        free_qrv_item(qi);
    }
}


void free_qrv(struct qrvdb *qrvdb){
    clear_qrv(qrvdb);

    g_free(qrvdb->cur);
    g_hash_table_destroy(qrvdb->hash);
    g_free(qrvdb);
}


void qrv_add(gchar *call, gchar *wwl, int qrv_int, int wkd_me, const gchar *text){
    struct qrv_item *qi, *q;
    int ret;
    char raw[20];
     

    if (!wwl || !*wwl){
        wwl = find_wwl_by_call(cw, call);
        if (!wwl) return;
    }
    //dbg("qrv_add('%s', '%s', %d, %d, '%s')\n", call, wwl, qrv_int, wkd_me, text);
    
    foreach(q, qrvdb->list) {
        ret = strcasecmp(call, q->call);
        if (ret==0){ // found same, update info
            if (wwl && *wwl) {
                CONDGFREE(q->wwl);
                q->wwl = uc(g_strdup(wwl)); 
                qrv_compute_qrbqtf(q);
            }
            //q->bands |= qrv_int;  why comment ???
            if (text && strlen(text) > strlen(q->text)){
                CONDGFREE(q->text);
                q->text = g_strdup(text);
            }
            return;
        }
        if (ret<=0) break; 
    }
    
    qi = g_new0(struct qrv_item, 1);
    qi->call  = uc(g_strdup(call));
    qi->wwl   = uc(g_strdup(wwl));
    qi->bands_qrv = qrv_int;
    qi->bands_wkd_me = wkd_me;
    qi->text  = g_strdup(text);
    qrv_compute_qrbqtf(qi);

    add_at_pos(q->prev, qi);
    get_raw_call(raw, call);
    g_hash_table_insert(qrvdb->hash, g_strdup(raw), qi); 
    //printf("add_to_list(%s %s %d %s)\r\n", qi->call, qi->wwl, qi->bands, qi->text);
}

int is_qrv(char *s){
    char *c, d, last = '\0';

    if (!s) return 0;
    if (strlen(s)==0) return 0;
    for (c=s; *c!='\0'; c++){
        d = toupper(*c);
        if (d < 'A' || d > 'Z') return 0;
        if (d < last) return 0;
        last = d;
    }
    return 1;
}

#define QRV_DELIM " \t\r\n"
void load_one_qrv(struct qrvdb *qrvdb, gchar *s){
    gchar *call, *wwl, *qrv_str, *text;
    int qrv_int, free_text = 0, bands_wkd_me;
    char *token_ptr, *c;

    //dbg("load_one_qrv('%s')\n", s);
    call = strtok_r(s, QRV_DELIM, &token_ptr);
    if (!call) return;

    wwl = strtok_r(NULL, QRV_DELIM, &token_ptr);
    if (!wwl) return;

    qrv_str = strtok_r(NULL, QRV_DELIM, &token_ptr);
    if (!qrv_str) qrv_str = "";

    text = token_ptr;

    if (!is_qrv(qrv_str)){
        qrv_int = qrvdb->def_bands;
        bands_wkd_me = qrvdb->def_bands;
        text = g_strconcat(qrv_str, " ", text, NULL);
        free_text = 1;
    }else{
        qrv_int = 0;
        bands_wkd_me = 0;
        for (c=qrv_str;*c!='\0';c++){
            if (isupper(*c)) bands_wkd_me |= 1 << (*c - 'A');
            *c=upcase(*c);
            qrv_int|=1<<(*c-'A');
        }
    }

    qrv_add(call, wwl, qrv_int, bands_wkd_me, text);
    if (free_text) g_free(text);

}

int load_qrv_from_file(struct qrvdb *qrvdb, gchar *filename){
    FILE *f;
    char s[202];

    f = fopen(filename, "rt");
    if (!f){
        return -1;
    }

    while((fgets(s, 200, f))!=NULL){
        chomp(s);
        load_one_qrv(qrvdb,s);
    }
    fclose(f);
    return 0;
}


int save_qrv_to_file(const char *filename){
    FILE *f;
    struct qrv_item *qi;
    char qrv_str[32];
    int i;
    char *c;

    f = fopen(filename, "wt");
    if (!f){
        log_addf(VTEXT(T_CANT_WRITE_S), filename);
        return errno;
    }

    foreach(qi, qrvdb->list) {
        for (i=0, c=qrv_str;i<32;i++){
            if (!(qi->bands_qrv & (1<<i))) continue;
            if (qi->bands_wkd_me & (1<<i))
                *c='A'+i;
            else
                *c='a'+i;
            c++;
        }
        *c='\0';
        fprintf(f, "%s\t%s\t%s\t%s\n", qi->call, qi->wwl, qrv_str, qi->text); 
    }

    fclose(f);
    return 0;
}

int qrv_skip(struct qrv_item *qi, int bi){
    //if (get_qso_by_callsign(aband, qi->call)!=NULL) return 1;
//    dbg("qrv_skip(%s)\n", qi->call);
    if ((qi->bands_qrv & (1<<bi))==0) return 1;
    if ((qi->bands_wkd & (1<<bi))!=0) return 1;
    return 0;
}

void qrv_seek(struct qrvdb *qrvdb, int value, int whence){
    int bi;
    struct qrv_item *qi = NULL;
                
    if (!aband) return;
    bi=lowcase(aband->bandchar)-'a';

   // dbg("\nqrv_seek(%d, %d)\n", value, whence);

    if (whence==1) qrvdb->cur[bi]=qrvdb->list.next;
    if (whence==-1) qrvdb->cur[bi]=qrvdb->list.prev;


    if (value<0) {
        for(qi = qrvdb->cur[bi]; value < 0; ){
            qi = qi->prev;
            //printf("try %s\n", qi->call);
            if (qi == &qrvdb->list) return;
            if (qrv_skip(qi, bi)) continue;
            value++;
            qrvdb->cur[bi] = qi;
        }
        return;
    }
    
    if (value>0) {
        for(qi = qrvdb->cur[bi]; value > 0; ){
            qi = qi->next;
            //printf("try %s\n", qi->call);
            if (qi == &qrvdb->list) return;
            if (qrv_skip(qi, bi)) continue;
            value--;
            qrvdb->cur[bi] = qi;
        }
        return;
    }
}


int sw_qrv_kbd_func(struct subwin *sw, struct event *ev, int fw){
    int bi;
    
    if (!aband) return 0;
    bi=lowcase(aband->bandchar)-'a';
    
    switch(kbd_action(KM_MAIN,ev)){
        case ACT_ESC:
            return 0;
            break;
        case ACT_DOWN:
            qrv_seek(qrvdb, 1, 0);
            redraw_later();
            return 1;
        case ACT_UP:
            qrv_seek(qrvdb, -1, 0);
            redraw_later();
            return 1;
        case ACT_PAGE_DOWN:
            qrv_seek(qrvdb, sw->h-1, 0);
            redraw_later();
            return 1;
        case ACT_PAGE_UP:
            qrv_seek(qrvdb, -(sw->h-1), 0);
            redraw_later();
            return 1;
        case ACT_HOME:
            qrv_seek(qrvdb, 1, 1); /* seek to end */
            redraw_later();
            return 1;
        case ACT_END:
            qrv_seek(qrvdb, 1, -1); /* seek to end */
            redraw_later();
            return 1;
		case ACT_SCROLL_LEFT:
    		if (sw->ho>0) sw->ho--;
            redraw_later();
            return 1;
		case ACT_SCROLL_RIGHT:
			sw->ho++;
            redraw_later();
            return 1;
        case ACT_ENTER:
            sw_unset_focus();
            il_set_focus(aband->il);
            if (qrvdb->cur[bi]){
                clear_tmpqsos(aband,1);
                process_input(aband, qrvdb->cur[bi]->call, 1);
                redraw_later();
/*#ifdef HAVE_HAMLIB                
                if (trig){
                    trig_set_qrg(trig, qrvdb->cur[bi]->qrg*1000.0);
                }
#endif                */
            }
            return 1;
        case ACT_DELETE:
            if (!qrvdb->cur[bi]) break;
            qrvdb->cur[bi]->bands_qrv &= ~(1<<bi);
            qrv_seek(qrvdb, 1, 0);
            redraw_later();
            return 1;
        default:
            if (ev->x==' ') {
                if (qrvdb->cur[bi]){
                    clear_tmpqsos(aband,1);
                    process_input(aband, qrvdb->cur[bi]->call, 0);
                    redraw_later();
                }
            }
    }
    return 0;
}

int sw_qrv_mouse_func(struct subwin *sw, struct event *ev, int fw){
    return 0;
}

void sw_qrv_draw_qrv(struct subwin *sw, int i, struct qrv_item *qi, int color){
    gchar *c;

    if (!qi || !qi->call){
        c = g_strdup("~");
    }else{
        c = g_strdup_printf("%-12s  %6s  %s", qi->call, qi->wwl, qi->text);
    }

    if (c && strlen(c)>sw->ho){
        print_text(sw->x, sw->y+i, sw->w, c + sw->ho, color);
    }
    g_free(c);
}


void sw_qrv_redraw(struct subwin *sw){
    struct qrv_item *qi;
    int i,bi;

    if (!aband) return;
    
    bi=lowcase(aband->bandchar)-'a';
    fill_area(sw->x, sw->y+sw->h/2, sw->w, 1, COL_INV);
    sw_qrv_draw_qrv(sw, sw->h/2, qrvdb->cur[bi], COL_INV);
   
    if (!qrvdb->cur[bi]) qrv_seek(qrvdb, 1, 1);

    if (qrvdb->cur[bi]) qi=qrvdb->cur[bi]->prev;
    else              qi=NULL;
    
    for (i=sw->h/2-1;i>=0; i--){
        if (!qi) break;
        if (qi==&qrvdb->list) break;
        if (qrv_skip(qi, bi))
            i++;
        else
            sw_qrv_draw_qrv(sw, i, qi, COL_NORM);
        
        qi=qi->prev;
    }
    
    if (qrvdb->cur[bi]) qi=qrvdb->cur[bi]->next;
    else              qi=NULL;

    for (i=sw->h/2+1;i<sw->h; i++){
        if (!qi) break;
        if (qi==&qrvdb->list) break;
        if (qrv_skip(qi, bi))
            i--;
        else
                sw_qrv_draw_qrv(sw, i, qi, COL_NORM);
        qi=qi->next;
    }
}

void sw_qrv_check_bounds(struct subwin *sw){
}
    
void qrv_compute_qrbqtf(struct qrv_item *q){
    double qtf, qtfrad;

    if (!q) return;
    /* CHANGE look at add_tmpqso_locator */    
    qrbqtf(ctest->pwwlo, q->wwl, &q->qrb, &qtf, NULL, 2);
    q->qtf = (int) (qtf+0.5);
    if (q->qrb < 0.1) {
        q->qrb = 0;
        q->qtf = 0;
    } 
    qtfrad=qtf*MY_PI/180.0;
    q->kx =   q->qrb*sin(qtfrad);
    q->ky = - q->qrb*cos(qtfrad);

    if (q->qrb < 600){
        q->weight = q->qrb;
    }else{
        q->weight= 1800 - 2 * q->qrb;
        if (q->weight < 0) q->weight = 0;
    }
    /*dbg("compute_qrbqtf(%s) %s->%s qrb=%f kx=%d ky=%d\n", 
            q->call, ctest->pwwlo, q->wwl, q->qrb, q->kx, q->ky); */
}


void qrv_recalc_qrbqtf(struct qrvdb *qrvdb){
    struct qrv_item *q;
    foreach(q, qrvdb->list) {
        qrv_compute_qrbqtf(q);
    }

}

void qrv_recalc_gst(struct qrvdb *qrvdb){
    struct qrv_item *q;
    int i, j, max;
    int qtf2;

    if (!qrvdb) return;
                    
    qrvdb->beamwidth = 10;
    for (i=0;i<=180;i++){
        qrvdb->antchar[i]=cos(MY_PI/180.0*(double)i*120.0/qrvdb->beamwidth);
       /* dbg("i=%3d %5.3f\n", i, antchar[i]);*/
        if (qrvdb->antchar[i]<=0) break;
    }
    for (;i<=180;i++) qrvdb->antchar[i]=0;
    
    for (i=0; i<360; i++) qrvdb->gst[i]=0.0;
    foreach(q, qrvdb->list) {
        if (q->qtf<=0 || q->qtf>=360) continue;
        
        if (aband && qrv_skip(q, aband->bandchar-'a')) continue;
        
        for (j=0;j<=180;j++){
            qtf2 = q->qtf + j;
            if (qrvdb->antchar[j] == 0) break;
            if (qtf2<0) qtf2 += 360;
            qrvdb->gst[qtf2%360] += q->weight * qrvdb->antchar[j];
        }
        for (j=1;j<180;j++){
            qtf2 = q->qtf - j;
            if (qrvdb->antchar[j] == 0) break;
            if (qtf2<0) qtf2 += 360;
            qrvdb->gst[qtf2%360] += q->weight * qrvdb->antchar[j];
        }

    }
    max = 0;
    for (i=0; i<360; i++) if (qrvdb->gst[i]>max) max = qrvdb->gst[i];
    if (!max) return;
    for (i=0; i<360; i++) qrvdb->gst[i] = qrvdb->gst[i] / max; 
} 


void qrv_set_wkd(struct qrvdb *qrvdb, struct qso *qso){
    char raw[20];
    struct qrv_item *qi;

    if (!qso) return;
    if (!qso->band) return;
            
    get_raw_call(raw, qso->callsign);
    qi = (struct qrv_item *)g_hash_table_lookup(qrvdb->hash, raw);

    int bi = qso->band->bandchar - 'a';
    if (!qi) return;
    qi->bands_wkd |= 1 << bi;
}

void qrv_recalc_wkd(struct qrvdb *qrvdb){
    struct qrv_item *qi;
    struct band *b;
    struct qso *qso;
    int i, j;
    char raw[20];

    ST_START;

    foreach (qi, qrvdb->list){
        qi->bands_wkd = 0;
    }

    if (!ctest) return;
    for (i=0; i<ctest->bands->len; i++){
        b = (struct band *)g_ptr_array_index(ctest->bands, i);
        for (j=0; j<b->qsos->len; j++){
            qso = (struct qso *)g_ptr_array_index(b->qsos, j);
            if (qso->dupe) continue;
            if (qso->error) continue;
            get_raw_call(raw, qso->callsign);
            qi = (struct qrv_item *)g_hash_table_lookup(qrvdb->hash, raw);
            //dbg("recalc_qrv_wkd(%s) qi=%p\n", raw, qi);
            if (!qi) continue;
            qi->bands_wkd |= 1 << (b->bandchar - 'a');
        }
    }

    ST_STOP;

}
